package ru.dbotthepony.kstarbound.render

import ru.dbotthepony.kstarbound.gl.*
import ru.dbotthepony.kstarbound.math.FloatMatrix
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.ITileChunk
import ru.dbotthepony.kstarbound.world.World
import kotlin.collections.ArrayList

class ChunkRenderer(val state: GLStateTracker, val chunk: Chunk, val world: World? = null) : AutoCloseable {
	private val foregroundLayers = TileLayerList()
	private val backgroundLayers = TileLayerList()
	private val bakedMeshes = ArrayList<BakedStaticMesh>()
	private val unloadableBakedMeshes = ArrayList<BakedStaticMesh>()

	private fun getForeground(): ITileChunk {
		return world?.getForegroundView(chunk.pos) ?: chunk.foreground
	}

	private fun getBackground(): ITileChunk {
		return world?.getBackgroundView(chunk.pos) ?: chunk.background
	}

	/**
	 * Тесселирует "статичную" геометрию в builders (к примеру тайлы).
	 *
	 * Может быть вызван вне рендер потока (ибо в любом случае он требует некой "стаитичности" данных в чанке)
	 * но только если до этого был вызыван loadRenderers() и геометрия чанка не поменялась
	 */
	fun tesselateStatic() {
		if (state.isSameThread()) {
			for (mesh in bakedMeshes) {
				mesh.close()
			}

			bakedMeshes.clear()
		} else {
			unloadableBakedMeshes.addAll(bakedMeshes)
			bakedMeshes.clear()
		}

		foregroundLayers.clear()

		val foreground = getForeground()

		// TODO: Синхронизация (ибо обновления игровой логики будут в потоке вне рендер потока)
		for ((pos, tile) in foreground.posToTile) {
			if (tile != null) {
				val renderer = state.tileRenderers.get(tile.def.materialName)
				renderer.tesselate(foreground, foregroundLayers, pos)
			}
		}

		backgroundLayers.clear()

		val background = getBackground()

		for ((pos, tile) in background.posToTile) {
			if (tile != null) {
				val renderer = state.tileRenderers.get(tile.def.materialName)
				renderer.tesselate(background, backgroundLayers, pos, background = true)
			}
		}
	}

	/**
	 * Принудительно подгружает в GLStateTracker все необходимые рендереры (ибо им нужны текстуры и прочее)
	 *
	 * Вызывается перед tesselateStatic()
	 */
	fun loadRenderers() {
		unloadUnused()

		// TODO: Синхронизация (ибо обновления игровой логики будут в потоке вне рендер потока)
		for ((_, tile) in getForeground().posToTile) {
			if (tile != null) {
				state.tileRenderers.get(tile.def.materialName)
			}
		}

		for ((_, tile) in getBackground().posToTile) {
			if (tile != null) {
				state.tileRenderers.get(tile.def.materialName)
			}
		}
	}

	private fun unloadUnused() {
		if (unloadableBakedMeshes.size != 0) {
			for (baked in unloadableBakedMeshes) {
				baked.close()
			}

			unloadableBakedMeshes.clear()
		}
	}

	fun uploadStatic(clear: Boolean = true) {
		unloadUnused()

		for ((baked, builder) in backgroundLayers.buildList()) {
			bakedMeshes.add(BakedStaticMesh(baked, builder))
		}

		for ((baked, builder) in foregroundLayers.buildList()) {
			bakedMeshes.add(BakedStaticMesh(baked, builder))
		}

		if (clear) {
			backgroundLayers.clear()
			foregroundLayers.clear()
		}
	}

	fun render(transform: FloatMatrix<*> = state.matrixStack.last) {
		unloadUnused()

		for (mesh in bakedMeshes) {
			mesh.render(transform)
		}
	}

	override fun close() {
		for (mesh in bakedMeshes) {
			mesh.close()
		}

		for (mesh in unloadableBakedMeshes) {
			mesh.close()
		}
	}
}
